
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: envoy.code.check:package

from setuptools import setup

setup(**{
    'author': 'Ryan Northey',
    'author_email': 'ryan@synca.io',
    'description': '"Code checker used in Envoy proxy\'s CI"',
    'entry_points': {
        'console_scripts': [
            'envoy.code.check=envoy.code.check:run',
        ],
    },
    'install_requires': (
        'abstracts>=0.0.12',
        'aio.core>=0.9.1',
        'aio.run.checker>=0.5.7',
        'envoy.base.utils>=0.3.9',
        'flake8>=5.0.4',
        'packaging>=23.0',
        'pyyaml',
        'yamllint',
        'yapf',
    ),
    'license': 'Apache Software License 2.0',
    'long_description': """
envoy.code.check
================

Code checker used in Envoy proxy's CI
""",
    'maintainer': 'Ryan Northey',
    'maintainer_email': 'ryan@synca.io',
    'name': 'envoy.code.check',
    'namespace_packages': (
    ),
    'package_data': {
        'envoy.code.check': (
            'py.typed',
        ),
    },
    'packages': (
        'envoy.code.check',
        'envoy.code.check.abstract',
    ),
    'url': 'https://github.com/envoyproxy/pytooling/tree/main/envoy.code.check',
    'version': '0.3.6',
})
